con
   NOTAG     = $FFFF
   NOFID     = $FFFFFFFF

   #100, t_version, r_version, t_auth, r_auth, t_attach, r_attach
   t_error, r_error, t_flush, r_flush, t_walk, r_walk, t_open, r_open
   t_create, r_create, t_read, r_read, t_clunk, r_clunk

'' maximum length we're willing to send/receive from host
'' write: 4 + 1 + 2 + 4 + 8 + 4 + 1024 = 1048
   buffSize      = 1048

   FS_MODE_READ  = 0
   FS_MODE_WRITE = 1
   FS_MODE_TRUNC = 16

   #0, fSendRecv
var
   long maxlen
   byte txbuf[buffSize]

pub doPut1(ptr,x) : result             ' Put one byte in host buffer
   byte[ptr++] := x
   return ptr

pub doPut2(ptr,x) : result             ' Put one word in host buffer
   repeat 2                            '  Order is LSB -> MSB
      byte[ptr++] := x & $FF
      x >>= 8
   return ptr

pub doPut4(ptr,x)                      ' Put one long in host buffer
   repeat 4                            '  Order is LSB -> MSB
      byte[ptr++] := x & $FF
      x >>= 8
   return ptr

pub doPutStr(ptr,s)                    ' Put contents of string in host buffer
   repeat strsize(s)
      byte[ptr++] := byte[s++]
   return ptr

pub fetch1(ptr) : result               ' Get one byte from host buffer
   result := byte[ptr++]

pub fetch2(ptr) : result               ' Get one word from host buffer
   repeat 2
      result := (result << 8) | byte[ptr++]

pub fetch4(ptr) : result               ' Get one long from host buffer
   repeat 4
      result := (result << 8) | byte[ptr++]

// initialize connection to host
// returns 0 on success, -1 on failure
// "fn" is the function to send a 9P protocol request to
// the host and read back a reply
pub fs_init(func) : result | ptr,size,len,msize,s,tag ' Initialize connection
   maxlen := buffSize
   ptr := doPut4(@txbuf,0)
   ptr := doPut1(ptr,t_version)
   ptr := doPut2(ptr,NOTAG)
   ptr := doPut4(ptr,MAXLEN)
   ptr := doPutStr(ptr,string("9P2000"))
   case func
      fSendRecv: sendRecv(@txbuf,ptr,maxlen)
   ptr := @txbuf + 4
   if ptr[0] != r_version
      return -1
   tag := fetch2(ptr + 1)
   msize := fetch4(ptr + 3)   
   s := fetch2(ptr + 7)
   if s <> 6 or 0 <> strcomp(@ptr[9],string("9P2000"))
      outputFormat(string("Bad version response from host: s=%d ver=%s\n"))
      outputItem(s)
      outputItem(ptr+9)
      outputFlush(nullPtr)
      return -1
   if msize < 64 or msize > MAXLEN
      outputFormat(string("Max message size %d is out of range\n"))
      outputItem(msize)
      outputFlush(nullPtr)
      return -1
   maxlen := msize
    // OK, try to attach
    ptr := doPut4(txbuf, 0);  // space for size
    ptr := doPut1(ptr, t_attach);
    ptr := doPut2(ptr, NOTAG);  // not sure about this one...
    ptr := doPut4(ptr, (uint32_t)&rootdir); // our FID
    ptr := doPut4(ptr, NOFID); // no authorization requested
    ptr := doPutStr(ptr, "user");
    ptr := doPutStr(ptr, ""); // no aname
    case func
       fSendRecv: len := sendRecv(@txbuf, ptr, maxlen);
    ptr := @txbuf+4;
    if byte[ptr] <> r_attach
       outputFormat(string("Unable to attach\n"))
       outputFlush(nullPtr)
       return -1;
    return 0;

fs_file *fs_getroot()
{
    return &rootdir;
}

// walk from fid "dir" along path, creating fid "newfile"
// if "skipLast" is nonzero, then do not try to walk the last element
// (needed for create or other operations where the file may not exist)
static int do_fs_walk(fs_file *dir, fs_file *newfile, const char *path, int skipLast)
{
    uint8_t *ptr;
    uint8_t *sizeptr;
    int c;
    uint32_t curdir = (uint32_t) dir;
    int len;
    int r;
    int sentone = 0;
    
    do {
        ptr = doPut4(txbuf, 0); // space for size
        ptr = doPut1(ptr, t_walk);
        ptr = doPut2(ptr, NOTAG);
        ptr = doPut4(ptr, curdir);
        curdir = (uint32_t)newfile;
        ptr = doPut4(ptr, curdir);
        while (*path == '/') path++;
        len = ptr - txbuf;
        if (*path) {
            ptr = doPut2(ptr, 1); // nwname
            sizeptr = ptr;
            ptr = doPut2(ptr, 0);
            while (*path && *path != '/' && len < maxlen) {
                *ptr++ = *path++;
                len++;
            }
	    if (skipLast && !*path) {
	      if (sentone) {
		return 0;
	      } else {
		doPut2(sizeptr, 1);
		ptr = sizeptr+2;
		ptr = doPut1(ptr, '.');
	      }
	    } else {
	      doPut2(sizeptr, (uint32_t)(ptr - (sizeptr+2)));
	    }
        } else {
            ptr = doPut2(ptr, 0);
        }

        r = (*sendRecv)(txbuf, ptr, maxlen);
	sentone = 1;
        if (txbuf[4] != r_walk) {
            return -1;
        }
    } while (*path);
    return 0;
}

fs_walk(dirPtr,newfilePtr,pathPtr) : result
   return do_fs_walk(dirPtr, newfilePtr, pathPtr, 0)

fs_open_relative(dirPtr,fPtr,pathPtr,fs_mode) : result | r, ptr, mode
    mode := fs_mode
    r := fs_walk(dirPtr,fPtr,pathPtr)
    if r <> 0
       return r
    ptr := doPut4(txbuf,0);  '' space for size
    ptr := doPut1(ptr,t_open)
    ptr := doPut2(ptr,NOTAG)
    ptr := doPut4(ptr,fPtr)
    ptr := doPut1(ptr,mode)
    r := sendRecv(@txbuf,ptr,maxlen)
    if txbuf[4] <> r_open
       return -1
    long[fPtr][0] := long[fPtr][1] := 0
    return 0

int fs_open(fs_file *f, const char *path, int fs_mode)
{
    return fs_open_relative(&rootdir, f, path, fs_mode);
}

int fs_create(fs_file *f, const char *path)
{
    int r;
    fs_file dir;
    uint8_t *ptr;
    uint32_t permissions = 0666; // rw for all
    const char *lastname;
    
    // first, walk to the directory containing our target file
    r = do_fs_walk(&rootdir, f, path, 1);
    if (r < 0) {
        // directory not found
        return -2;
    }
    lastname = strrchr(path, '/');
    if (!lastname) {
        lastname = path;
    } else {
        lastname++;
    }
    // try to create the file
    ptr = doPut4(txbuf, 0); // space for size
    ptr = doPut1(ptr, t_create);
    ptr = doPut2(ptr, NOTAG);
    ptr = doPut4(ptr, (uint32_t)f);
    ptr = doPutStr(ptr, lastname);
    ptr = doPut4(ptr, permissions);
    ptr = doPut1(ptr, FS_MODE_TRUNC | FS_MODE_WRITE);
    r = (*sendRecv)(txbuf, ptr, maxlen);
    if (r >= 0 && txbuf[4] == r_create) {
      return r;
    }
    
    // can we open the file with truncation set?
    r = fs_open_relative(f, f, lastname, FS_MODE_TRUNC | FS_MODE_WRITE);
    if (r < 0) {
        return r;
    }
    return 0;
}

int fs_close(fs_file *f)
{
    uint8_t *ptr;
    int r;
    ptr = doPut4(txbuf, 0); // space for size
    ptr = doPut1(ptr, t_clunk);
    ptr = doPut2(ptr, NOTAG);
    ptr = doPut4(ptr, (uint32_t)f);
    r = (*sendRecv)(txbuf, ptr, maxlen);
    if (r < 0 || txbuf[4] != r_clunk) {
        return -1;
    }
    return 0;
}

int fs_read(fs_file *f, uint8_t *buf, int count)
{
    uint8_t *ptr;
    int totalread = 0;
    int curcount;
    int r;
    int left;
    uint32_t oldlo;
    while (count > 0) {
        ptr = doPut4(txbuf, 0); // space for size
        ptr = doPut1(ptr, t_read);
        ptr = doPut2(ptr, NOTAG);
        ptr = doPut4(ptr, (uint32_t)f);
        ptr = doPut4(ptr, f->offlo);
        ptr = doPut4(ptr, f->offhi);
        left = maxlen - (ptr + 4 - txbuf);
        if (count < left) {
            curcount = count;
        } else {
            curcount = left;
        }
        ptr = doPut4(ptr, curcount);
        r = (*sendRecv)(txbuf, ptr, maxlen);
        if (r < 0) return r;
        ptr = txbuf + 4;
        if (*ptr++ != r_read) {
            return -1;
        }
        ptr += 2; // skip tag
        r = FETCH4(ptr); ptr += 4;
        if (r < 0 || r > curcount) {
            return -1;
        }
        if (r == 0) {
            // EOF reached
            break;
        }
        memcpy(buf, ptr, r);
        buf += r;
        totalread += r;
        count -= r;
        oldlo = f->offlo;
        f->offlo = oldlo + r;
        if (f->offlo < oldlo) {
            f->offhi++;
        }
    }
    return totalread;
}

int fs_write(fs_file *f, const uint8_t *buf, int count)
{
    uint8_t *ptr;
    int totalread = 0;
    int curcount;
    int r;
    int left;
    uint32_t oldlo;
    while (count > 0) {
        ptr = doPut4(txbuf, 0); // space for size
        ptr = doPut1(ptr, t_write);
        ptr = doPut2(ptr, NOTAG);
        ptr = doPut4(ptr, (uint32_t)f);
        ptr = doPut4(ptr, f->offlo);
        ptr = doPut4(ptr, f->offhi);
        left = maxlen - (ptr + 4 - txbuf);
        if (count < left) {
            curcount = count;
        } else {
            curcount = left;
        }
        ptr = doPut4(ptr, curcount);
        // now copy in the data
        memcpy(ptr, buf, curcount);
        r = (*sendRecv)(txbuf, ptr+curcount, maxlen);
        if (r < 0) return r;
        ptr = txbuf + 4;
        if (*ptr++ != r_write) {
            return -1;
        }
        ptr += 2; // skip tag
        r = FETCH4(ptr); ptr += 4;
        if (r < 0 || r > curcount) {
            return -1;
        }
        if (r == 0) {
            // EOF reached
            break;
        }
        buf += r;
        totalread += r;
        count -= r;
        oldlo = f->offlo;
        f->offlo = oldlo + r;
        if (f->offlo < oldlo) {
            f->offhi++;
        }
    }
    return totalread;
}
